#Module ExtensionMethod.psm1
#
#Fonctions d'aide  la cration de fichier ps1xml ddi aux 
# mthodes d'extension contenues dans un fichier assembly dotnet.
#
#D'aprs une ide de Bart De Smet's :
# http://bartdesmet.net/blogs/bart/archive/2007/09/06/extension-methods-in-windows-powershell.aspx 
#
#Laurent Dardenne
# v1.0 le 31/07/2010
#exemple :
#   $TypesFileName="DateTimeExtensionMethod.Types.ps1xml" 
#   Add-Type -Path "$PathAsm\Developpez.Dotnet.dll" -Pass|
#       #filtre sur le namespace
#     #Where {$_.Namespace -eq "Developpez.Dotnet.Reflection"} |
#    Find-ExtensionMethod -ExcludeGeneric|
#    Get-ExtensionMethodInfo -ExcludeGeneric -ExcludeInterface|
#    New-HashTable -key "Key" -Value "Value" -MakeArray|
#     #Cre un texte reprsentant une structure XML que 
#     # l'on insre dans un fichier de type PowerShell.
#    New-ExtensionMethodTypeData > $TypesFileName
#    
#   Type $TypesFileName|More 
#    #Si la stratgie d'excution est 'AllSigned' 
#    if (Get-ExcecutionPolicy -eq "AllSigned" )
#     {
#       #Signe $TypesFileName
#     }
#    #Met  jour la configuration de type tendu (ETS)
#   Update-TypeData $TypesFile
#   

function Find-ExtensionMethod{
 param( 
    [ValidateNotNull()]
    [Parameter(Position=0, Mandatory=$true,ValueFromPipeline = $true)]
   [System.Type] $Type,
   [switch] $ExcludeGeneric
 ) 
 #Recherche et renvoi les mthodes d'extension contenu par le type $Types.
  #
  #On ne traite que les mthodes statiques publiques des types qui sont sclls, non-gnrique et non-imbriqus.
  # Spec C# 3.0 :
  #   Lorsque le premier paramtre d'une mthode inclut le modificateur this, la mthode est dite mthode d'extension. 
  #   Les mthodes d'extension ne peuvent tre dclares que dans des classes statiques non gnriques non imbriques. 
  #   Le premier paramtre d'une mthode d'extension ne peut pas comporter de modificateurs autres que this, et le type 
  #   de paramtre ne peut tre un type pointeur.  
  #
  # $ExcludeGeneric :  Ne renvoi pas les mthodes d'extension des types gnriques, car ces types ncessiteraient,
  #                    dans le fichier ps1xml, une dclaration pour chaque type utilis lors du paramtrage de la classe : 
  #                     MyClass<String>, MyClass<Int>, etc
  #
  # Add-Type -path "$Path\$AssemblyName.dll" -Pass|GetExtensionMethods -ExcludeGeneric
  
  process {
     #Filtre les types publics qui sont sclls, non-gnrique et non-imbriqus. 
     #Les classes statiques possdent les attributs abstract et Sealed, ce dernier est plac sur 
     # le type lors de la compilation (il n'est pas dclar dans le code source)
   $Type|
    where {$_.IsPublic -and $_.IsSealed -and $_.IsAbstract -and ($_.IsGenericType -eq $false) -and ($_.IsNested -eq $false)}| 
    Foreach {
        #recherche uniquement les mthodes statique publiques
       $_.GetMethods(([System.Reflection.BindingFlags]"Static,Public"))|
         #Filtre les mthodes d'extension.
         # Note: en C# l'attribut ExtensionAttribute est dfini par le compilateur.
        Where {$_.IsDefined([System.Runtime.CompilerServices.ExtensionAttribute], $false)}|
        Foreach { 
           #on renvoi toutes les mthodes
          if ($ExcludeGeneric -eq $False) 
           {$_}
           #on renvoi toutes les mthodes sauf les mthodes gnriques.
          elseif ($_.IsGenericMethod -eq $false)
           {$_}
       }
    }
  }
}

function Test-ClassExtensionMethods([System.Type] $Type) 
{ #Dtermine si le type $Type contient des mthodes d'extension.
  #
  @($Type|
    where {$_.IsPublic -and $_.IsSealed -and $_.IsAbstract -and ($_.IsGenericType -eq $false) -and ($_.IsNested -eq $false)}| 
    Foreach {
       $_.GetMethods(([System.Reflection.BindingFlags]"Static,Public"))|
        Where {$_.IsDefined([System.Runtime.CompilerServices.ExtensionAttribute], $false)}|
    }).count -gt 0
}

function Get-ExtensionMethodInfo{
  Param (
     [ValidateNotNull()]
     [Parameter(Position=0, Mandatory=$true,ValueFromPipeline = $true)] 
    [System.Reflection.MethodInfo] $MethodInfo,
    [switch] $ExcludeGeneric,
    [switch] $ExcludeInterface
  )
   
  #On renvoi,  partir d'informations d'une mthode d'extension, un objet DictionaryEntry dont :
  #       la cl est le type du premier paramtre, dclar dans la signature de la mthode (le type peut tre une type imbriqu),
  #       la valeur est une mthode d'extension rattache  ce mme type.
  #
  # $ExcludeGeneric :  Ne renvoi pas les mthodes d'extension dont le type du premier paramtre est un type gnrique, 
  #                    car le nom de la mthode ncessiterait, dans le fichier ps1xml, une dclaration pour chaque type 
  #                    utilis lors du paramtrage de la mthode : 
  #                      MyClass.MaMethod<DateTimefull strong name ...>, MyClass.MaMethod<Double full strong name ...>, etc
  #
  # $ExcludeInterface : Ne renvoi pas les mthodes d'extension dont le type du premier paramtre est un type interface.
  #                     PowerShell v2 ne sait pas 'extraire' une interface particulire  partir d'un objet.
  #                     De plus les objets de type IEnumerable sont transforms en System.Array par PowerShell...
 process {
     #Le premier paramtre de la mthode dtermine le type sur lequel on dclare la mthode d'extension.
     #En C# c'est le rle du mot cl 'this'.
     #Une mthode d'extension a au moins 1 paramtre.
    $Parameter=($MethodInfo.GetParameters())[0]
    if (!($Parameter.ParameterType.IsInterface -and $ExcludeInterface)) 
    {
       #Si -ExcludeGeneric est prcis on ne traite pas les types gnriques
       #
       #ContainsGenericParameters : 
       # renvoie true si l'objet Type est lui-mme un paramtre de type gnrique ou a des paramtres 
       # de type pour lesquels les types spcifiques n'ont pas t fournis; sinon, false.
      if (!( $ExcludeGeneric -and 
           ($Parameter.ParameterType.IsGenericType -or 
            $Parameter.ParameterType.ContainsGenericParameters 
           )
         ))
      {  
         #Pas d'exclusion demande ou le type n'est pas gnrique.
         #On renvoi une paire cl/valeur avec le type du premier paramtre et l'objet mthode
        new-object System.Collections.DictionaryEntry(
                                         #la cl est le nom du type du premier paramtre de l'objet mthode
                                       ($Parameter.ParameterType.ToString()),
                                         #la valeur est l'objet mthode
                                        $_
                                      )
      } #if ExcludeGeneric
    } #if ExcludeInterface  
 }#process
}

Function New-HashTable {
#From http://blogs.msdn.com/b/powershell/archive/2007/11/27/new-hashtable.aspx 
# Author:   laurent Dardenne (ajout $Value) 
# Version:  0.2 
# Author:   Jeffrey Snover 
# Version:  0.1 
#Requires -Version 1.0 
param( 
    $Key=$(Throw "USAGE: New-HashTable -Key <property>"),
    $Value,  
    [Switch]$NoOverWrite, 
    [Switch]$MakeArray 
    ) 
  Begin 
  { 
      $hash = @{} 
  } 
  Process 
  {  
       #On dfinit la cl
      $Property = $_.$key
      
       #On dfinit la valeur de la cl  partir
       # du nom de proprit contenue dans $Value 
      if ([String]::IsNullOrEmpty($Value))
      {  $Object=$_
      }else 
      {  $Object=$_.$Value
      }
       
       #Pas d'crasement de la cl si elle existe 
      if ($NoOverWrite -And $hash.$Property) 
      {  Write-Error "$Property already exists" 
      }elseif ($MakeArray) #Il existe plusieurs occurences d'une mme cl, 
      {                    # on cre un tableau afin de mmoriser toutes les valeurs 
         if (!$hash.$Property) 
         {   $hash.$Property = new-object System.Collections.ArrayList 
         } 
         [void]$hash.$Property.Add($Object) 
      }else 
      {    #Il n'existe qu'une occurence d'une cl
           #On la remplace si elle existe.
         $hash.$Property = $Object 
      } 
  } 
  End 
  { 
      $hash 
  }
}

function Format-TableExtensionMethod{
 #Affiche sur la console la liste des mthodes d'extension regroupes par type.
 #Wrapper spcialis du cmdlet Format-Table.  
 param( 
    [ValidateNotNull()]
    [Parameter(Position=0, Mandatory=$true,ValueFromPipeline = $true)]
    [System.Collections.HashTable] $Hashtable
 )
#
#Hashtable : Hashtable contenant les mthodes d'extension.
$Hashtable.GetEnumerator()|
 Foreach {
   $InstanceType=$_.Key;$_.Value|
    Sort $_.ParameterCount|
    Add-Member NoteProperty InstanceType $InstanceType -pass
 }|
 Format-Table @{Label="Method";Expression={ $_.ToString()}} -GroupBy InstanceType
}


Function New-ExtensionMethodTypeData{
#Renvoi un texte reprsentant une structure XML  insrer dans un fichier .ps1xml
 param( 
    [ValidateNotNull()]
    [Parameter(Position=0, Mandatory=$true,ValueFromPipeline = $true)]
  [System.Collections.HashTable] $Hashtable
 ) 
#$Hashtable : Hashtable contenant les mthodes d'extension.

#Pour chaque type, on cre autant de balise <ScriptMethod> que de mthodes recenses.
#Comme chaque mthode peut tre surcharge,  on doit considrer le type de la surcharge, 
# par le nombre de paramtres et par leurs types :
#
#  1- public static string To(this SubStringFrom subStringFrom)
#  2- public static string To(this SubStringFrom subStringFrom, int end)
#  3- public static string To(this SubStringFrom subStringFrom, string end)
#  4- public static string To(this SubStringFrom subStringFrom, string end, bool includeBoundary)
#
#On teste le nombre de paramtres pour gnrer la balise script qui contient le code PowerShell :
#         <Script>
# 						switch ($args.Count) {
#            			0 { [Developpez.Dotnet.StringExtensions]::To($this)}
# 	 							1 { [Developpez.Dotnet.StringExtensions]::To($this,$args[0])}
# 	 							2 { [Developpez.Dotnet.StringExtensions]::To($this,$args[0] ,$args[1])}
# 	          default { throw "La mthode To ne propose pas de signature contenant $($args.Count) paramtres." }
#           }
#         </Script>   
#
# Dans cet exemple, pour $args.Count = 1 (on ne compte pas $this) on ne doit gnrer qu'une seule ligne, 
#on laisse le shell invoquer la mthode.
#Si le nombre de paramtres correspond, mais pas leurs type alors le shell dclenchera une exception. 

Write-Output @"
<?xml version="1.0" encoding="utf-8"?>
 <Types>
"@
 
 #Pour chaque type (classe/struct)
$Hashtable.GetEnumerator()| 
 Foreach {
   Write-Output @"  
   <Type>
    <Name>$($_.Key)</Name>
    <Members>
"@
  $_.Value| Group Name |
    #Pour chaque groupe de mthodes du type courant
    #Utilise une Here-String imbriques
  Foreach {
@"
      <ScriptMethod>
        <Name>$($_.Name)</Name>
        <Script>
$("`t"*6)switch (`$args.Count) {
           $(
             $_.group|
                #Pour faciliter le tri on ajoute un membre contenant le nombre de paramtre de la mthode.
              Add-member ScriptProperty ParameterCount -value {$this.GetParameters().Count} -pass|
                #On ne cre qu'une seule ligne pour les mthodes surcharges utilisant un nombre identique de paramtres.
                #Exemple :
                #  public static SubStringFrom From(this string s, int start);
                #  public static SubStringFrom From(this string s, string start);
                #Dans ce cas leurs types est diffrent, ce qui ne pose pas de pb car PowerShell est non typ.
                #
              Sort ParameterCount -unique|
              foreach {
               $Count=$_.GetParameters().Count
                 #On soustrait 1  $Count pour crer un dcalage :
                 #    0 =$this                   -> $Objet.Method()
                 #    1=$this,$args[0]           -> $Objet.Method($Param1)
                 #    2=$this,$args[0],$args[1]  -> $Objet.Method($Param1,$Param2)
                 #
                 # Le dcalage sur le nombre de paramtres est d au fait que l'on doit prendre en charge 
                 # le modificateur C# this, qui est gal  $this dans un script de mthode PowerShell, celui-ci
                 # n'est pas prcis lors de l'appel de la mthode PowerShell.
                 #
                 #On construit plusieurs lignes respectant la syntaxe d'appel d'une mthode d'extension( mthode statique) :
                 #  nb_argument  { [TypeName]::MethodName($this, 0..n arguments) }  
@"
$("`t"*7)$($Count-1) { [$($_.Declaringtype)]::$($_.Name)(`$this$(
              "$(if ($Count -gt 1){1..($Count-1)|Foreach {",`$args[$($_-1)]"}})"))}`r`n`t
"@
              }
           )
            default { throw "La mthode $($_.Name) ne propose pas de signature contenant `$(`$args.Count) paramtres." }
          }
        </Script>
      </ScriptMethod>
"@
} 
Write-Output @"  
     </Members>
   </Type>
"@
}
Write-Output @"
 </Types>
"@ 
}
